{*******************************************************}
{File:      NCOciBuff.PAS                               }
{Revision:  0.04.02 / 23.04.2000                        }
{Comment:   NC OCI8 VCL: scrollable cursor implement.   }
{Copyright: (c) 1999-2000, Dmitry Arefiev               }
{Author:    Dmitry Arefiev, dmitrya@inthink.com         }
{*******************************************************}
{$I NCOciDef.inc}

unit NCOciBuff;

interface

Uses Classes, Forms, SysUtils, NCOciWrapper, NCOci;

type
    TOCICursor = class;
    TOCIDeadCursor = class;

    // ------------------------------------------------------------------------
    // Generic cursor
    // ------------------------------------------------------------------------

    TOCIRecordStatus = (rsEmpty, rsFetched, rsOld, rsNewInserted, rsNewUpdated,
        rsNewDeleted, rsNewInsApplyed, rsNewUpdApplyed, rsNewDelApplyed,
        rsDelphiBuff);
    POCIRecordStatus = ^TOCIRecordStatus;
    TOCIRecordStatuses = set of TOCIRecordStatus;

    TOCICursorFilter = procedure (ASender: TOCICursor; var AAccept: Boolean) of object;
    TOCIUpdateRecordEvent = procedure (Sender: TOCICursor; var AAction: TOCIUpdateAction) of object;
    TOCIUpdateErrorEvent = procedure (Sender: TOCICursor; E: Exception; var AAction: TOCIUpdateAction) of object;

    POCIBookmark = pointer;

    TOCICursorRecProp = record
        FStatus: TOCIRecordStatus;
        FNumber: sb4;
    end;
    POCICursorRecProp = ^TOCICursorRecProp;

    TOCICursor = class
    private
        function GetDefVar(APosition: sb4): TOCIVariable;
        function GetDefVarCount: ub4;
        procedure SetCachedUpdates(const Value: Boolean);
    protected
        FDefVars: TList;
        FEOF, FBOF: Boolean;
        FOnFilterRecord: TOCICursorFilter;
        FCursorCanceled: Boolean;
        FCachedUpdates: Boolean;
        FInApplyUpdates: Boolean;
        FUpdatesPending: Boolean;
        FStatusFilter: TOCIRecordStatuses;
        FOnUpdateRecord: TOCIUpdateRecordEvent;
        FOnUpdateError: TOCIUpdateErrorEvent;
        procedure CheckActive;
        procedure CheckInactive;
        function GetRecNo: sb4; virtual; abstract;
        procedure SetRecNo(AValue: sb4); virtual; abstract;
        function GetActive: Boolean; virtual; abstract;
        function GetRecordCount: sb4; virtual; abstract;
        function GetRecordStatus: TOCIRecordStatus; virtual; abstract;
        function GetBookmark: POCIBookmark; virtual; abstract;
        procedure GotoBookmark(ABmk: POCIBookmark); virtual; abstract;
        function GetBookmarkSize: ub4; virtual; abstract;
        function GetRecordSize: ub4;
        procedure SetCurrentRecord(P: POCIBookmark; AResync: Boolean); virtual; abstract;
        procedure CheckBidir; virtual; abstract;
        procedure CheckNoBLOB;
    public
        constructor Create;
        destructor Destroy; override;
        procedure AddDefVar(AVar: TOCIVariable);
        procedure RemoveDefVar(APosition: sb4);

        function GetData(ABmk: POCIBookmark; APosition: ub4; ABuff: pUb1;
            var ASize: sb4): Boolean; virtual; abstract;
        function GetDataPtr(ABmk: POCIBookmark; APosition: ub4; var ABuff: pUb1;
            var ASize: sb4): Boolean; virtual; abstract;
        procedure SetData(ABmk: POCIBookmark; APosition: ub4; ABuff: pUb1;
            ASize: sb4); virtual; abstract;
        procedure SetIsNull(ABmk: POCIBookmark; APosition: ub4;
            AIsNull: Boolean); virtual; abstract; // for LOB fields

        function AllocateRecord: POCIBookmark; virtual; abstract;
        procedure FreeRecord(P: POCIBookmark); virtual; abstract;
        procedure InitRecord(P: POCIBookmark); virtual; abstract;
        procedure GetRecord(P, PRef: POCIBookmark; var ARecProc: TOCICursorRecProp); virtual; abstract;
        procedure DeleteRecord(pToDel: POCIBookmark); virtual; abstract;
        procedure ModifyRecord(pToUpd, P: POCIBookmark); virtual; abstract;
        procedure InsertRecord(P: POCIBookmark); virtual; abstract;
        function CompareBookmarks(P1, P2: POCIBookmark): sb2; virtual; abstract;
        function BookmarkValid(P: POCIBookmark): Boolean; virtual; abstract;
        procedure CopyRecord(PCopyFrom, PCopyTo: POCIBookmark); virtual; abstract;

        procedure OpenComplete; virtual; abstract;
        procedure Close; virtual; abstract;
        procedure Skip(ANum: sb4); virtual; abstract;
        procedure First(AtBOF: Boolean); virtual; abstract;
        procedure Last(AtEOF: Boolean); virtual; abstract;
        procedure FetchAll; virtual;

        function GetNewRecord(P: POCIBookmark): POCIBookmark; virtual; abstract;
        function GetOldRecord(P: POCIBookmark): POCIBookmark; virtual; abstract;
        procedure ApplyUpdates(P: POCIBookmark); virtual; abstract;
        procedure CancelUpdates(P: POCIBookmark); virtual; abstract;
        procedure CommitUpdates(P: POCIBookmark); virtual; abstract;
        procedure EndUpdate(P: POCIBookmark);

        property EOF: Boolean read FEOF;
        property BOF: Boolean read FBOF;
        property RecordSize: ub4 read GetRecordSize;
        property BookmarkSize: ub4 read GetBookmarkSize;
        property Bookmark: POCIBookmark read GetBookmark write GotoBookmark;
        property RecNo: sb4 read GetRecNo write SetRecNo;
        property RecordCount: sb4 read GetRecordCount;
        property RecordStatus: TOCIRecordStatus read GetRecordStatus;
        property Active: Boolean read GetActive;
        property DefVarCount: ub4 read GetDefVarCount;
        property DefVar[AIndex: sb4]: TOCIVariable read GetDefVar;

        property OnFilterRecord: TOCICursorFilter read FOnFilterRecord write FOnFilterRecord;
        property CursorCanceled: Boolean read FCursorCanceled;
        property CachedUpdates: Boolean read FCachedUpdates write SetCachedUpdates;
        property InApplyUpdates: Boolean read FInApplyUpdates;
        property UpdatesPending: Boolean read FUpdatesPending; 
        property StatusFilter: TOCIRecordStatuses read FStatusFilter write FStatusFilter;
        property OnUpdateRecord: TOCIUpdateRecordEvent read FOnUpdateRecord write FOnUpdateRecord;
        property OnUpdateError: TOCIUpdateErrorEvent read FOnUpdateError write FOnUpdateError;
    end;

    // ------------------------------------------------------------------------
    // Scrollable cursor with dead caching
    // ------------------------------------------------------------------------

    POCIDeadBookmark = ^TOCIDeadBookmark;
    PPOCIDeadBookmark = ^POCIDeadBookmark;
    TOCIDeadBookmark = record
        FPageIndex: sb4;
        FRecIndex: sb4;
    end;

    POCIDeadRecordHeader = ^TOCIDeadRecordHeader;
    PPOCIDeadRecordHeader = ^POCIDeadRecordHeader;
    TOCIDeadRecordHeader = record
        FSelfBmk: TOCIDeadBookmark;
        FStatus: TOCIRecordStatus;
        FRecordNumber: sb4;
        FRefRecord: POCIDeadRecordHeader;
    end;

    POCIDeadPage = pUb1;
    // TOCIDeadRecordHeader[0]
    // ...
    // TOCIDeadRecordHeader[FRecPerPage - 1]
    // DATA Var[1]
    // ...
    // DATA Var[FDefVars.Count]

    TUB4s = array[0..511] of ub4;
    PUB4s = ^TUB4s;

    TOCIDeadCursor = class(TOCICursor)
    private
        FStmt: TOCIStatement;
        FRecPerPage: sb4;
        FPageSize: sb4;
        FPages: TList;
        FPosition: TOCIDeadBookmark;
        FRowsInserted, FRowsDeleted: ub4;
        FVarOffsets: PUB4s;
        FFreeRecordsPage: sb4;
        FUnidirectional: Boolean;
        FOpenComplete: Boolean;
        FDisableFetch: Boolean;
        procedure PageUpdateSize;
        procedure PageClear(APageInd: ub4);
        function PageNew: Integer;
        procedure PageFree(APageInd: ub4);
        function PageRecordHeader(APageInd, ARecInd: ub4): POCIDeadRecordHeader;
        procedure PageBind(APageInd: ub4);
        procedure PageFetched(APageInd: ub4);
        procedure FreePages;
        function PageRecordData(APageInd: ub4; APosition: ub4): pUb1;
        procedure ValidatePosition(APageInd, ARecInd: sb4; AExcludeEO: Boolean);
{$IFDEF OCI_DEBUG}
        function P2RecHead(P: Pointer): POCIDeadRecordHeader;
{$ENDIF}        
        procedure ClearRecord(PRec: POCIDeadBookmark);
        function AllocateRecordInternal(AStatus: TOCIRecordStatus): POCIDeadRecordHeader;
        procedure CheckModifable;
        procedure UpdateFreeRecPage(APageInd: ub4);
    protected
        function GetRecNo: sb4; override;
        procedure SetRecNo(AValue: sb4); override;
        function GetActive: Boolean; override;
        function GetRecordCount: sb4; override;
        function GetBookmark: POCIBookmark; override;
        procedure GotoBookmark(ABmk: POCIBookmark); override;
        function GetBookmarkSize: ub4; override;
        function GetRecordStatus: TOCIRecordStatus; override;
        function SetRecordStatus(APageInd, ARecInd: ub4; AStatus: TOCIRecordStatus): POCIDeadRecordHeader;
        procedure SetCurrentRecord(P: POCIBookmark; AResync: Boolean); override;
        procedure CheckBidir; override;
    public
        constructor Create;
        destructor Destroy; override;

        function GetData(ABmk: POCIBookmark; APosition: ub4; ABuff: pUb1;
            var ASize: sb4): Boolean; override;
        function GetDataPtr(ABmk: POCIBookmark; APosition: ub4; var ABuff: pUb1;
            var ASize: sb4): Boolean; override;
        procedure SetData(ABmk: POCIBookmark; APosition: ub4; ABuff: pUb1;
            ASize: sb4); override;
        procedure SetIsNull(ABmk: POCIBookmark; APosition: ub4;
            AIsNull: Boolean); override; // for LOB fields

        function AllocateRecord: POCIBookmark; override;
        procedure FreeRecord(P: POCIBookmark); override;
        procedure InitRecord(P: POCIBookmark); override;
        procedure GetRecord(P, PRef: POCIBookmark; var ARecProc: TOCICursorRecProp); override;
        procedure DeleteRecord(pToDel: POCIBookmark); override;
        procedure ModifyRecord(pToUpd, P: POCIBookmark); override;
        procedure InsertRecord(P: POCIBookmark); override;
        function CompareBookmarks(P1, P2: POCIBookmark): sb2; override;
        function BookmarkValid(P: POCIBookmark): Boolean; override;
        procedure CopyRecord(PCopyFrom, PCopyTo: POCIBookmark); override;

        procedure Open(AStmt: TOCIStatement; AService: TOCIService;
            AFetchSize: sb4; AExactRows: sb4; AUnidirectional: Boolean);
        procedure OpenComplete; override;
        procedure Describe(AStmt: TOCIStatement; AService: TOCIService);
        procedure Close; override;
        procedure Skip(ANum: sb4); override;
        procedure First(AtBOF: Boolean); override;
        procedure Last(AtEOF: Boolean); override;
        procedure FetchAll; override;

        function GetNewRecord(P: POCIBookmark): POCIBookmark; override;
        function GetOldRecord(P: POCIBookmark): POCIBookmark; override;
        procedure ApplyUpdates(P: POCIBookmark); override;
        procedure CancelUpdates(P: POCIBookmark); override;
        procedure CommitUpdates(P: POCIBookmark); override;
    end;

implementation

Uses NCOciMsg;

// ------------------------------------------------------------------------
// Generic cursor
// ------------------------------------------------------------------------

constructor TOCICursor.Create;
begin
    inherited Create;
    FDefVars := TList.Create;
    FStatusFilter := [rsFetched, rsNewInserted, rsNewUpdated,
        rsNewInsApplyed, rsNewUpdApplyed];
end;

destructor TOCICursor.Destroy;
begin
    FDefVars.Free;
    inherited Destroy;
end;

procedure TOCICursor.AddDefVar(AVar: TOCIVariable);
begin
    CheckInactive;
    while FDefVars.Count < AVar.Position do
        FDefVars.Add(nil);
    FDefVars[AVar.Position - 1] := AVar;
end;

procedure TOCICursor.CheckActive;
    procedure BuffMBOpen;
    begin
        OCIDBError(msgOCIBuffMBOpen);
    end;
begin
    if not Active then
        BuffMBOpen;
end;

procedure TOCICursor.CheckInactive;
    procedure BuffMBClose;
    begin
        OCIDBError(msgOCIBuffMBClose);
    end;
begin
    if Active then
        BuffMBClose;
end;

function TOCICursor.GetDefVarCount: ub4;
begin
    Result := FDefVars.Count;
end;

function TOCICursor.GetDefVar(APosition: sb4): TOCIVariable;
begin
    if (APosition < 1) or (APosition > FDefVars.Count) then
        OCIDBError(msgOCIInvalidVarIndex);
    Result := TOCIVariable(FDefVars[APosition - 1]);
end;

procedure TOCICursor.RemoveDefVar(APosition: sb4);
begin
    FDefVars.Delete(APosition - 1);
end;

function TOCICursor.GetRecordSize: ub4;
var
    i: Integer;
begin
    Result := 0;
    for i := 0 to FDefVars.Count - 1 do
        with TOCIVariable(FDefVars[i]) do
            Inc(Result, BindRecLen);
end;

procedure TOCICursor.EndUpdate(P: POCIBookmark);
begin
    FUpdatesPending := True;
    if not CachedUpdates then
        ApplyUpdates(P)
    else
        SetCurrentRecord(P, True);
end;

procedure TOCICursor.FetchAll;
begin
    Last(False);
end;

procedure TOCICursor.SetCachedUpdates(const Value: Boolean);
begin
    if FCachedUpdates <> Value then begin
        if not Value then
            CancelUpdates(nil)
        else begin
            CheckBidir;
            CheckNoBLOB;
        end;
        FCachedUpdates := Value;
    end;
end;

procedure TOCICursor.CheckNoBLOB;
var
    i: Integer;
begin
    for i := 1 to DefVarCount do
        if DefVar[i].DataType in [otCursor, otCLOB, otBLOB, otCFile, otBFile] then
            OCIDBError(msgOCIBadDT4CU);
end;

// ------------------------------------------------------------------------
// Scrollable cursor with dead caching
// ------------------------------------------------------------------------

constructor TOCIDeadCursor.Create;
begin
    inherited Create;
    FPages := TList.Create;
    FFreeRecordsPage := -1;
end;

destructor TOCIDeadCursor.Destroy;
begin
    if Active then
        Close;
    FPages.Free;
    inherited Destroy;
end;

function TOCIDeadCursor.GetActive: Boolean;
begin
    Result := FStmt <> nil;
end;

procedure TOCIDeadCursor.UpdateFreeRecPage(APageInd: ub4);
begin
    if (FRecPerPage > 1) and
       ((FFreeRecordsPage = -1) or (ub4(FFreeRecordsPage) > APageInd)) then
        FFreeRecordsPage := APageInd;
end;

procedure TOCIDeadCursor.PageUpdateSize;
var
    i: Integer;
begin
    FPageSize := 0;
    GetMem(FVarOffsets, FDefVars.Count * SizeOf(Integer));
    for i := 0 to FDefVars.Count - 1 do
        with TOCIVariable(FDefVars[i]) do begin
            ArrayLen := FRecPerPage;
            FVarOffsets[i] := FPageSize;
            FPageSize := FPageSize + BindRecLen * FRecPerPage;
        end;
    FPageSize := FPageSize + ((SizeOf(TOCIDeadRecordHeader) + 3) and not 3) * FRecPerPage;
end;

function TOCIDeadCursor.SetRecordStatus(APageInd, ARecInd: ub4; AStatus: TOCIRecordStatus): POCIDeadRecordHeader;
const
    statWithHndl: TOCIRecordStatuses = [rsFetched, rsOld, rsNewInserted,
        rsNewInsApplyed, rsNewDeleted, rsNewDelApplyed];

    procedure ForEachVar(AInit: Boolean; ANoHandleAlloc: Boolean);
    var
        i: Integer;
    begin
        for i := 0 to FDefVars.Count - 1 do
            with TOCIVariable(FDefVars[i]) do begin
                ArrayLen := FRecPerPage;
                ExternBuffer := PageRecordData(APageInd, i + 1);
                if AInit then
                    InitBuffer(ARecInd, ANoHandleAlloc)
                else
                    ClearBuffer(ARecInd, ANoHandleAlloc);
            end;
    end;

begin
    Result := PageRecordHeader(APageInd, ARecInd);
    with Result^ do begin
        if ((FStatus in statWithHndl) and not (AStatus in statWithHndl)) or
           ((FStatus <> rsEmpty) and (AStatus = rsEmpty))  then
            ForEachVar(False, not (FStatus in statWithHndl))
        else if not (FStatus in statWithHndl) and (AStatus in statWithHndl) then
            ForEachVar(True, False);
        FStatus := AStatus;
    end;
end;

procedure TOCIDeadCursor.PageClear(APageInd: ub4);
var
    i: ub4;
begin
    for i := 0 to FRecPerPage - 1 do
        SetRecordStatus(APageInd, i, rsEmpty);
end;

function TOCIDeadCursor.PageNew: Integer;
begin
    Result := FPages.Add(AllocMem(FPageSize));
{$IFDEF OCI_DEBUG}
    PageClear(Result);
{$ENDIF}    
end;

procedure TOCIDeadCursor.PageFree(APageInd: ub4);
var
    p: POCIDeadRecordHeader;
begin
    try
        PageClear(APageInd);
    finally
        p := POCIDeadRecordHeader(FPages[APageInd]);
        FreeMem(p, FPageSize);
        FPages.Delete(APageInd);
    end;
end;

procedure TOCIDeadCursor.FreePages;
begin
    while FPages.Count > 0 do
        PageFree(0);
    FreeMem(FVarOffsets);
end;

function TOCIDeadCursor.PageRecordHeader(APageInd, ARecInd: ub4): POCIDeadRecordHeader;
begin
{$IFDEF OCI_DEBUG}
    ValidatePosition(APageInd, ARecInd, True);
{$ENDIF}
    Result := POCIDeadRecordHeader(LongWord(FPages[APageInd]) +
        SizeOf(TOCIDeadRecordHeader) * ARecInd);
end;

function TOCIDeadCursor.PageRecordData(APageInd: ub4; APosition: ub4): pUb1;
begin
{$IFDEF OCI_DEBUG}
    ValidatePosition(APageInd, FRecPerPage - 1, True);
{$ENDIF}
    Result := pUb1(ub4(FPages[APageInd]) +
        ub4(SizeOf(TOCIDeadRecordHeader) * FRecPerPage) +
        FVarOffsets[APosition - 1]);
end;

procedure TOCIDeadCursor.PageBind(APageInd: ub4);
var
    i: Integer;
begin
    for i := 0 to FRecPerPage - 1 do
        SetRecordStatus(APageInd, i, rsFetched);
    for i := 0 to FDefVars.Count - 1 do
        TOCIVariable(FDefVars[i]).Bind;
end;

procedure TOCIDeadCursor.PageFetched(APageInd: ub4);
var
    i: ub4;
    rNo: ub4;
begin
    FCursorCanceled := FStmt.EOF;
    if (FStmt.LastRowCount = 0) then begin
        if not FUnidirectional then
            PageFree(APageInd);
    end
    else begin
        rNo := RecordCount - FStmt.LastRowCount + 1;
        for i := 0 to FStmt.LastRowCount - 1 do
            with PageRecordHeader(APageInd, i)^ do begin
                FSelfBmk.FPageIndex := APageInd;
                FSelfBmk.FRecIndex := i;
                FRecordNumber := rNo + i;
                FRefRecord := nil;
            end;
        for i := FStmt.LastRowCount to FRecPerPage - 1 do
            SetRecordStatus(APageInd, i, rsEmpty);
        if FStmt.LastRowCount <= FRecPerPage - 1 then
            UpdateFreeRecPage(APageInd);
    end;
end;

procedure TOCIDeadCursor.Open(AStmt: TOCIStatement; AService: TOCIService;
    AFetchSize: sb4; AExactRows: sb4; AUnidirectional{, AToFirst}: Boolean);
begin
    CheckInactive;
    if AFetchSize < 1 then
        AFetchSize := 1;
    if AExactRows > 0 then
        AFetchSize := AExactRows;
    with FPosition do begin
        FPageIndex := -1;
        FRecIndex := AFetchSize - 1;
    end;
    FRecPerPage := AFetchSize;
    FCursorCanceled := False;
    FEOF := False;
    FBOF := False;
    FRowsInserted := 0;
    FRowsDeleted := 0;
    FStmt := AStmt;
    FUnidirectional := AUnidirectional;
    FFreeRecordsPage := -1;
    try
        PageUpdateSize;
        PageNew;
        PageBind(0);
        FStmt.Execute(AService, AFetchSize, AExactRows > 0, False, False);
    except
        FreePages;
        FStmt := nil;
        FCursorCanceled := True;
        raise;
    end;
    PageFetched(0);
    if not AUnidirectional and (AExactRows < 0) then
        FetchAll;
{    if AToFirst then begin
        Skip(1);
        FBOF := FEOF;
    end;
}end;

procedure TOCIDeadCursor.Describe(AStmt: TOCIStatement; AService: TOCIService);
begin
    FStmt := AStmt;
    FStmt.Describe(AService);
end;

procedure TOCIDeadCursor.Close;
begin
{$IFDEF OCI_DEBUG}
    CheckActive;
{$ENDIF}
    FreePages;
    try
        if not FCursorCanceled then begin
            FCursorCanceled := True;
            FStmt.CancelCursor;
        end;
    finally
        FStmt := nil;
    end;
end;

procedure TOCIDeadCursor.CheckBidir;
begin
    if FUnidirectional and FOpenComplete then
        OCIDBError(msgOCIMustBeBidir);
end;

procedure TOCIDeadCursor.OpenComplete;
begin
    FOpenComplete := True;
end;

procedure TOCIDeadCursor.Skip(ANum: sb4);
var
    d: Integer;
    actualNo: ub4;
    i: Integer;
    ok, unidir: Boolean;
begin
{$IFDEF OCI_DEBUG}
    CheckActive;
{$ENDIF}
    if ANum = 0 then begin
        Skip(1);
        Skip(-1);
        Exit;
    end;
    with FPosition do begin
        if ANum < 0 then begin
            actualNo := -ANum;
            d := -1;
            FBOF := (FPageIndex = -1);
            FEOF := False;
        end
        else begin
            actualNo := ANum;
            d := 1;
            FBOF := False;
            FEOF := (FPageIndex = FPages.Count) and FCursorCanceled;
        end;
        unidir := FUnidirectional and (FStatusFilter <> [rsEmpty]);
        while (actualNo > 0) and not FEOF and not FBOF do begin
            FRecIndex := FRecIndex + d;
            if FRecIndex = -1 then begin
                FPageIndex := FPageIndex - 1;
                FRecIndex := FRecPerPage - 1;
                if FPageIndex = -1 then
                    FBOF := True;
            end
            else if FRecIndex = FRecPerPage then begin
                if unidir then begin
                    if (FPageIndex <> -1) then
                        if FCursorCanceled then
                            FEOF := True
                        else begin
                            PageClear(0);
                            PageBind(0);
                            FStmt.Next(FRecPerPage);
                            PageFetched(0);
                            if FCursorCanceled then
                                FEOF := FStmt.LastRowCount = 0;
                        end;
                    FPageIndex := 0;
                    FRecIndex := 0;
                end
                else begin
                    FPageIndex := FPageIndex + 1;
                    FRecIndex := 0;
                    if FPageIndex = FPages.Count then
                        if FCursorCanceled or (FStatusFilter = [rsEmpty]) then
                            FEOF := True
                        else if FDisableFetch then
                            FEOF := True
                        else begin
                            i := PageNew;
                            try
                                PageBind(i);
                                FStmt.Next(FRecPerPage);
                                PageFetched(i);
                            except
                                PageFree(i);
                                raise;
                            end;
                            if FCursorCanceled then
                                FEOF := FStmt.LastRowCount = 0;
                        end;
                end;
            end;
            if not FEOF and not FBOF and
               (PageRecordHeader(FPageIndex, FRecIndex)^.FStatus in FStatusFilter) then
                if Assigned(FOnFilterRecord) then begin
                    ok := True;
                    FOnFilterRecord(Self, ok);
                    if ok then
                        Dec(actualNo);
                end
                else
                    Dec(actualNo);
        end;
    end;
end;

procedure TOCIDeadCursor.FetchAll;
var
    i: Integer;
begin
{$IFDEF OCI_DEBUG}
    CheckActive;
{$ENDIF}
    CheckBidir;
    i := -1;
    try
        while not FCursorCanceled do begin
            i := -1;
            i := PageNew;
            PageBind(i);
            FStmt.Next(FRecPerPage);
            PageFetched(i);
        end;
    except
        if i <> -1 then
            PageFree(i);
        raise;
    end;
end;

procedure TOCIDeadCursor.First(AtBOF: Boolean);
begin
{$IFDEF OCI_DEBUG}
    CheckActive;
{$ENDIF}
    CheckBidir;
    with FPosition do begin
        FPageIndex := -1;
        FRecIndex := FRecPerPage - 1;
    end;
    if not AtBOF then
        Skip(1);
    FBOF := FEOF;
end;

procedure TOCIDeadCursor.Last(AtEOF: Boolean);
begin
{$IFDEF OCI_DEBUG}
    CheckActive;
{$ENDIF}
    CheckBidir;
    FetchAll;
    with FPosition do begin
        FPageIndex := FPages.Count;
        FRecIndex := 0;
    end;
    if not AtEOF then
        Skip(-1);
    FEOF := FBOF;
end;

function TOCIDeadCursor.GetBookmarkSize: ub4;
begin
    Result := SizeOf(TOCIDeadBookmark);
end;

procedure TOCIDeadCursor.ValidatePosition(APageInd, ARecInd: sb4; AExcludeEO: Boolean);
    procedure InvBmk;
    begin
        OCIDBError(msgOCIInvalidBmk);
    end;
begin
{$IFDEF OCI_DEBUG}
    CheckActive;
{$ENDIF}
    if AExcludeEO then begin
        if (APageInd < 0) or (APageInd > FPages.Count - 1) then
            InvBmk;
    end
    else begin
        if (APageInd < -1) or (APageInd > FPages.Count) then
            InvBmk;
    end;
    if (ARecInd < 0) or (ARecInd > FRecPerPage) then
        InvBmk;
end;

function TOCIDeadCursor.GetBookmark: POCIBookmark;
begin
{$IFDEF OCI_DEBUG}
    CheckActive;
{$ENDIF}
    Result := @(PageRecordHeader(FPosition.FPageIndex, FPosition.FRecIndex)^.FSelfBmk);
end;

procedure TOCIDeadCursor.GotoBookmark(ABmk: POCIBookmark);
begin
{$IFDEF OCI_DEBUG}
    CheckActive;
{$ENDIF}
    with POCIDeadBookmark(ABmk)^ do
        ValidatePosition(FPageIndex, FRecIndex, False);
    SetCurrentRecord(ABmk, True);
    if not FBOF and not FEOF then
        with FPosition, PageRecordHeader(FPageIndex, FRecIndex)^ do
            if (FRefRecord <> nil) and (FStatus = rsDelphiBuff) then
                FPosition := FRefRecord^.FSelfBmk;
end;

procedure TOCIDeadCursor.SetCurrentRecord(P: POCIBookmark; AResync: Boolean);
begin
    if P <> nil then begin
        FPosition := POCIDeadBookmark(P)^;
        FBOF := FPosition.FPageIndex = -1;
        FEOF := FPosition.FPageIndex = FPages.Count;
    end;
    if AResync then
        Skip(0);
end;

function TOCIDeadCursor.GetRecNo: sb4;
begin
{$IFDEF OCI_DEBUG}
    CheckActive;
{$ENDIF}
    if FEOF or FUnidirectional then
        Result := RecordCount
    else if FBOF then
        Result := 0
    else
        with FPosition do
            Result := PageRecordHeader(FPageIndex, FRecIndex)^.FRecordNumber;
end;

procedure TOCIDeadCursor.SetRecNo(AValue: sb4);
var
    pgNo, rNo: sb4;
begin
{$IFDEF OCI_DEBUG}
    CheckActive;
{$ENDIF}
    CheckBidir;
    pgNo := (AValue - 1) div FRecPerPage;
    rNo := (AValue - 1) mod FRecPerPage;
    if FCursorCanceled and (pgNo >= FPages.Count) or (pgNo < 0) then
        OCIDBError(msgOCIInvalidRNO);
    if pgNo < FPages.Count then begin
        FPosition.FPageIndex := pgNo;
        FPosition.FRecIndex := rNo;
    end
    else begin
        FPosition.FPageIndex := FPages.Count - 1;
        FPosition.FRecIndex := FRecPerPage - 1;
    end;
    FEOF := False;
    FBOF := False;
    while not EOF and (RecNo < AValue) do
        Skip(1);
end;

function TOCIDeadCursor.GetRecordCount: sb4;
begin
    Result := FStmt.ROW_COUNT + FRowsInserted - FRowsDeleted;
end;

function TOCIDeadCursor.GetRecordStatus: TOCIRecordStatus;
begin
{$IFDEF OCI_DEBUG}
    CheckActive;
{$ENDIF}
    if FEOF or FBOF then
        Result := rsEmpty
    else with FPosition do
        Result := PageRecordHeader(FPageIndex, FRecIndex)^.FStatus;
end;

function TOCIDeadCursor.GetData(ABmk: POCIBookmark; APosition: ub4;
    ABuff: pUb1; var ASize: sb4): Boolean;
begin
    if ABmk = nil then
        ABmk := @FPosition;
    with DefVar[APosition], POCIDeadBookmark(ABmk)^ do begin
        ExternBuffer := PageRecordData(FPageIndex, APosition);
        Result := GetData(FRecIndex, ABuff, ASize, dfDataSet);
    end;
end;

function TOCIDeadCursor.GetDataPtr(ABmk: POCIBookmark; APosition: ub4;
    var ABuff: pUb1; var ASize: sb4): Boolean;
begin
    if ABmk = nil then
        ABmk := @FPosition;
    with DefVar[APosition], POCIDeadBookmark(ABmk)^ do begin
        ExternBuffer := PageRecordData(FPageIndex, APosition);
        Result := GetDataPtr(FRecIndex, ABuff, ASize);
    end;
end;

procedure TOCIDeadCursor.SetData(ABmk: POCIBookmark; APosition: ub4;
    ABuff: pUb1; ASize: sb4);
begin
    if ABmk = nil then
        ABmk := @FPosition;
    with DefVar[APosition], POCIDeadBookmark(ABmk)^ do begin
        ExternBuffer := PageRecordData(FPageIndex, APosition);
        SetData(FRecIndex, ABuff, ASize, dfDataSet);
    end;
end;

procedure TOCIDeadCursor.SetIsNull(ABmk: POCIBookmark; APosition: ub4; AIsNull: Boolean);
begin
    if ABmk = nil then
        ABmk := @FPosition;
    with DefVar[APosition], POCIDeadBookmark(ABmk)^ do begin
        ExternBuffer := PageRecordData(FPageIndex, APosition);
        SetIsNull(FRecIndex, AIsNull);
    end;
end;

procedure TOCIDeadCursor.CopyRecord(PCopyFrom, PCopyTo: POCIBookmark);
var
    i: Integer;
    p: pUb1;
    l: sb4;
begin
    if PCopyFrom = nil then
        PCopyFrom := @FPosition;
    if PCopyTo = nil then
        PCopyTo := @FPosition;
    for i := 1 to FDefVars.Count do
        with DefVar[i] do begin
            with POCIDeadBookmark(PCopyFrom)^ do begin
                ExternBuffer := PageRecordData(FPageIndex, i);
                GetDataPtr(FRecIndex, p, l);
            end;
            with POCIDeadBookmark(PCopyTo)^ do begin
                ExternBuffer := PageRecordData(FPageIndex, i);
                SetData(FRecIndex, p, l, dfOCI);
            end;
        end;
end;

procedure TOCIDeadCursor.ClearRecord(PRec: POCIDeadBookmark);
var
    i: Integer;
begin
    for i := 1 to FDefVars.Count do
        SetData(PRec, i, nil, 0);
end;

// -----------------------------------------------------------------
// -----------------------------------------------------------------
// Delphi buffer management

{$IFDEF OCI_DEBUG}
function TOCIDeadCursor.P2RecHead(P: POCIBookmark): POCIDeadRecordHeader;
begin
    with POCIDeadBookmark(P)^ do
        Result := PageRecordHeader(FPageIndex, FRecIndex);
    if Result^.FStatus <> rsDelphiBuff then
        OCIDBError(msgOCIModifyNotDelphiBuff);
end;
{$ENDIF}

function TOCIDeadCursor.AllocateRecordInternal(AStatus: TOCIRecordStatus): POCIDeadRecordHeader;
var
    savePos: TOCIDeadBookmark;
    saveStFlt: TOCIRecordStatuses;
begin
{$IFDEF OCI_DEBUG}
    CheckActive;
{$ENDIF}
    savePos := FPosition;
    saveStFlt := FStatusFilter;
    StatusFilter := [rsEmpty];
    FDisableFetch := True;
    try
        // Free record page is unknown
        if FFreeRecordsPage = -1 then begin
            // If all records fetched then go to tail page
            // and try find free record
            if FCursorCanceled and (FPages.Count > 0) then begin
                FPosition.FPageIndex := FPages.Count - 1;
                FPosition.FRecIndex := 0;
                Skip(1);
            end
            // ... else allocate new empty page
            else
                FEOF := True;
        end
        // ... else try find empty record from FFreeRecordsPage page
        else begin
            FPosition.FPageIndex := FFreeRecordsPage - 1;
            FPosition.FRecIndex := FRecPerPage - 1;
            Skip(1);
        end;
        // Not found, then allocate page
        if FEOF then begin
            FPosition.FPageIndex := PageNew;
            FPosition.FRecIndex := 0;
        end;
        FFreeRecordsPage := FPosition.FPageIndex;
        Result := SetRecordStatus(FPosition.FPageIndex, FPosition.FRecIndex, AStatus);
        with Result^ do begin
            FSelfBmk := FPosition;
            FRecordNumber := FPosition.FPageIndex * FRecPerPage + FPosition.FRecIndex + 1;
            FRefRecord := nil;
        end;
    finally
        FPosition := savePos;
        FStatusFilter := saveStFlt;
        FDisableFetch := False;
    end;
end;

function TOCIDeadCursor.AllocateRecord: POCIBookmark;
begin
    Result := @(AllocateRecordInternal(rsDelphiBuff)^.FSelfBmk);
end;

procedure TOCIDeadCursor.FreeRecord(P: POCIBookmark);
begin
    with POCIDeadBookmark(p)^ do begin
        with SetRecordStatus(FPageIndex, FRecIndex, rsEmpty)^ do begin
            if FRefRecord <> nil then
                FRefRecord^.FRefRecord := nil;
            FRefRecord := nil;
        end;
        UpdateFreeRecPage(FPageIndex);
    end;
end;

procedure TOCIDeadCursor.InitRecord(P: POCIBookmark);
begin
{$IFDEF OCI_DEBUG}
    P2RecHead(P)^.FRefRecord := nil;
{$ELSE}
    with POCIDeadBookmark(P)^ do
        PageRecordHeader(FPageIndex, FRecIndex)^.FRefRecord := nil;
{$ENDIF}        
    ClearRecord(P);
end;

procedure TOCIDeadCursor.GetRecord(P, PRef: POCIBookmark; var ARecProc: TOCICursorRecProp);
var
    pRH: POCIDeadRecordHeader;
begin
    with FPosition do
        pRH := PageRecordHeader(FPageIndex, FRecIndex);
{$IFDEF OCI_DEBUG}
    P2RecHead(P)^.FRefRecord := pRH;
{$ELSE}
    with POCIDeadBookmark(P)^ do
        PageRecordHeader(FPageIndex, FRecIndex)^.FRefRecord := pRH;
{$ENDIF}
    with ARecProc do begin
        POCIDeadBookmark(PRef)^ := pRH^.FSelfBmk;
        FStatus := pRH^.FStatus;
        if FUnidirectional then
            FNumber := RecordCount
        else
            FNumber := pRH^.FRecordNumber;
    end;
//    CopyRecord(@FPosition, P);
// Now record is not copied into local Delphi buffer,
// until we start edit it. It is optimization kind of
// navigational methods. See TOCIDataSet.GetEditBookmark 
end;

procedure TOCIDeadCursor.DeleteRecord(pToDel: POCIBookmark);
var
    pRH: POCIDeadRecordHeader;
    UndoValue: TOCIRecordStatus;
    pUndoVar: POCIRecordStatus;

    procedure InitUndo;
    begin
        pUndoVar := nil;
    end;

    procedure SaveUndo(AStatusVar: POCIRecordStatus);
    begin
        pUndoVar := AStatusVar;
        if pUndoVar <> nil then
            UndoValue := pUndoVar^;
    end;

    procedure DoUndo;
    begin
        if pUndoVar <> nil then
            pUndoVar^ := UndoValue;
    end;

begin
    CheckModifable;
    if pToDel = nil then
        pToDel := @FPosition;
    with POCIDeadBookmark(pToDel)^ do
        pRH := PageRecordHeader(FPageIndex, FRecIndex);
    InitUndo;
    try
        with pRH^ do
            case FStatus of
            rsNewInserted:
                begin
                    Dec(FRowsInserted);
                    FreeRecord(@FSelfBmk);
                    Exit;
                end;
            rsNewDeleted, rsNewDelApplyed:
                Exit;
            rsFetched, rsNewUpdated, rsNewInsApplyed, rsNewUpdApplyed:
                begin
                    SaveUndo(@FStatus);
                    with FSelfBmk do
                        SetRecordStatus(FPageIndex, FRecIndex, rsNewDeleted);
                end;
            rsOld:
                begin
                    DeleteRecord(@FRefRecord^.FSelfBmk);
                    Exit;
                end;
            rsEmpty, rsDelphiBuff:
                OCIDBError(msgOCICantModifyBuff);
            end;
        EndUpdate(@pRH^.FSelfBmk);
        Inc(FRowsDeleted);
    except
        DoUndo;
        raise;
    end;
end;

procedure TOCIDeadCursor.ModifyRecord(pToUpd, P: POCIBookmark);
var
    pCur: POCIDeadRecordHeader;
    pRH: POCIDeadRecordHeader;
    pBmk: POCIDeadBookmark;
    UndoValue: TOCIRecordStatus;
    pUndoVar: POCIRecordStatus;
    UndoRefValue: POCIDeadRecordHeader;
    pUndoRefVar: PPOCIDeadRecordHeader;
    UndoFreeRec: POCIDeadRecordHeader;

    procedure InitUndo;
    begin
        pUndoVar := nil;
        pUndoRefVar := nil;
        UndoFreeRec := nil;
    end;

    procedure SaveUndo(AStatusVar: POCIRecordStatus;
        ARefRecord, AFreeRec: PPOCIDeadRecordHeader);
    begin
        pUndoVar := AStatusVar;
        if pUndoVar <> nil then
            UndoValue := pUndoVar^;
        pUndoRefVar := @ARefRecord;
        if pUndoRefVar <> nil then
            UndoRefValue := pUndoRefVar^;
        if AFreeRec <> nil then
            UndoFreeRec := AFreeRec^;
    end;

    procedure DoUndo;
    begin
        if pUndoVar <> nil then
            pUndoVar^ := UndoValue;
        if pUndoRefVar <> nil then
            pUndoRefVar^ := UndoRefValue;
        if UndoFreeRec <> nil then
            FreeRecord(@UndoFreeRec^.FSelfBmk);
    end;

begin
    CheckModifable;
    pRH := nil;
    if pToUpd = nil then
        pToUpd := @FPosition;
    with POCIDeadBookmark(pToUpd)^ do
        pCur := PageRecordHeader(FPageIndex, FRecIndex);
    InitUndo;
    try
        with pCur^ do
            case FStatus of
            rsFetched, rsNewDeleted, rsNewDelApplyed, rsNewInsApplyed:
                begin
                    pRH := AllocateRecordInternal(rsNewUpdated);
                    pRH^.FRefRecord := pCur;
                    pRH^.FRecordNumber := pCur^.FRecordNumber;
                    SaveUndo(@FStatus, @FRefRecord, @pRH);
                    with FSelfBmk do
                        SetRecordStatus(FPageIndex, FRecIndex, rsOld);
                    FRefRecord := pRH;
                end;
            rsOld:
                begin
                    ModifyRecord(@FRefRecord^.FSelfBmk, P);
                    Exit;
                end;
            rsNewUpdated, rsNewInserted:
                pRH := pCur;
            rsNewUpdApplyed:
                begin
                    SaveUndo(@FStatus, nil, nil);
                    with FSelfBmk do
                        pRH := SetRecordStatus(FPageIndex, FRecIndex, rsNewUpdated);
                    CopyRecord(@FSelfBmk, @FRefRecord^.FSelfBmk);
                end;
            rsEmpty, rsDelphiBuff:
                OCIDBError(msgOCICantModifyBuff);
            end;
        pBmk := @(pRH^.FSelfBmk);
        CopyRecord(P, pBmk);
        EndUpdate(pBmk);
    except
        DoUndo;
        raise;
    end;
end;

procedure TOCIDeadCursor.InsertRecord(P: POCIBookmark);
var
    pRH: POCIDeadRecordHeader;
    pBmk: POCIDeadBookmark;
begin
    CheckModifable;
    pRH := AllocateRecordInternal(rsNewInserted);
    pBmk := @pRH^.FSelfBmk;
    try
        CopyRecord(P, pBmk);
        EndUpdate(pBmk);
        Inc(FRowsInserted);
    except
        FreeRecord(pBmk);
        raise;
    end;
end;

function TOCIDeadCursor.CompareBookmarks(P1, P2: POCIBookmark): sb2;
begin
    if (POCIDeadBookmark(P1).FPageIndex < POCIDeadBookmark(P2).FPageIndex) or
       (POCIDeadBookmark(P1).FPageIndex = POCIDeadBookmark(P2).FPageIndex) and
        (POCIDeadBookmark(P1).FRecIndex < POCIDeadBookmark(P2).FRecIndex) then
        Result := -1
    else
        Result := 1;
end;

function TOCIDeadCursor.BookmarkValid(P: POCIBookmark): Boolean;
begin
    try
        with POCIDeadBookmark(P)^ do
            PageRecordHeader(FPageIndex, FRecIndex);
        Result := True;
    except
        Result := False;
    end;
end;

function TOCIDeadCursor.GetNewRecord(P: POCIBookmark): POCIBookmark;
begin
    if p = nil then
        p := @FPosition;
    Result := nil;
    with POCIDeadBookmark(p)^, PageRecordHeader(FPageIndex, FRecIndex)^ do begin
        case FStatus of
             rsFetched, rsNewInserted, rsNewUpdated, rsNewInsApplyed, rsNewUpdApplyed:
                Result := @FSelfBmk;
             rsOld, rsNewDeleted, rsNewDelApplyed:
                Result := @FRefRecord^.FSelfBmk;
             rsDelphiBuff:
                Result := GetOldRecord(@FRefRecord^.FSelfBmk);
        end;
    end;
end;

function TOCIDeadCursor.GetOldRecord(P: POCIBookmark): POCIBookmark;
begin
    if p = nil then
        p := @FPosition;
    Result := nil;
    with POCIDeadBookmark(p)^, PageRecordHeader(FPageIndex, FRecIndex)^ do begin
        case FStatus of
            rsNewInserted, rsNewUpdated, rsNewInsApplyed, rsNewUpdApplyed:
                Result := @FRefRecord^.FSelfBmk;
            rsFetched, rsOld, rsNewDeleted, rsNewDelApplyed:
                Result := @FSelfBmk;
            rsDelphiBuff:
                Result := GetOldRecord(@FRefRecord^.FSelfBmk);
        end;
    end;
end;

procedure TOCIDeadCursor.CheckModifable;
begin
    if InApplyUpdates then
        OCIDBError(msgOCIRecurseApply);
    if not Assigned(OnUpdateRecord) then
        OCIDBError(msgOCINoApplyCallback);
end;

procedure TOCIDeadCursor.ApplyUpdates(P: POCIBookmark);
var
    prevStatusFilter: TOCIRecordStatuses;
    updAction: TOCIUpdateAction;
    oldStatus, newStatus: TOCIRecordStatus;
    allApplyed: Boolean;
    lastPosBmk: TOCIDeadBookmark;
begin
    CheckModifable;
    prevStatusFilter := StatusFilter;
    StatusFilter := [rsNewInserted, rsNewUpdated, rsNewDeleted];
    FInApplyUpdates := True;
    FDisableFetch := True;
    allApplyed := True;
    lastPosBmk := FPosition;
    try
        if P = nil then
            First(False)
        else
            SetCurrentRecord(P, False);
        while not EOF do begin
            updAction := uaFail;
            repeat
                try
                    OnUpdateRecord(Self, updAction);
                except
                    on E: Exception do begin
                        if Assigned(OnUpdateError) then
                            OnUpdateError(Self, E, updAction);
                        if not Assigned(OnUpdateError) or (updAction = uaReRaise) then begin
                            CancelUpdates(P);
                            raise;
                        end;
                    end;
                end;
            until updAction <> uaRetry;
            case updAction of
            uaAbort:
                begin
                    CancelUpdates(P);
                    Abort;
                end;
            uaFail:
                begin
                    CancelUpdates(P);
                    OCIDBError(msgOCIApplyFailed);
                end;
            uaApplied:
                begin
                    with FPosition do
                        oldStatus := PageRecordHeader(FPageIndex, FRecIndex)^.FStatus;
                    newStatus := rsEmpty;
                    case oldStatus of
                        rsNewInserted: newStatus := rsNewInsApplyed;
                        rsNewUpdated: newStatus := rsNewUpdApplyed;
                        rsNewDeleted: newStatus := rsNewDelApplyed;
                    end;
                    with FPosition do
                        SetRecordStatus(FPageIndex, FRecIndex, newStatus);
                end;
            uaSkip:
                allApplyed := False;
            end;
            if P = nil then
                Skip(1)
            else
                Break;
        end;
        FUpdatesPending := not allApplyed;
    finally
        StatusFilter := prevStatusFilter;
        FInApplyUpdates := False;
        FDisableFetch := False;
        SetCurrentRecord(@lastPosBmk, True);
    end;
end;

procedure TOCIDeadCursor.CancelUpdates(P: POCIBookmark);
var
    prevStatusFilter: TOCIRecordStatuses;
    pRH: POCIDeadRecordHeader;
    toP: POCIBookmark;

    procedure SetFetched(pRH: POCIDeadRecordHeader);
    begin
        with pRH^, FSelfBmk do begin
            SetRecordStatus(FPageIndex, FRecIndex, rsFetched);
            if FRefRecord <> nil then
                FreeRecord(@FRefRecord^.FSelfBmk);
            toP := @FSelfBmk;
        end;
    end;

begin
    prevStatusFilter := StatusFilter;
    StatusFilter := [rsNewInserted, rsNewUpdated, rsNewDeleted];
    FDisableFetch := True;
    try
        if P = nil then
            First(False)
        else begin
            SetCurrentRecord(P, False);
            toP := nil;
        end;
        while not EOF do begin
            with FPosition do
                pRH := PageRecordHeader(FPageIndex, FRecIndex);
            with pRH^ do
                case FStatus of
                rsNewInserted:
                    begin
                        Dec(FRowsInserted);
                        FreeRecord(@FSelfBmk);
                    end;
                rsNewUpdated:
                    SetFetched(pRH^.FRefRecord);
                rsNewDeleted:
                    begin
                        Dec(FRowsDeleted);
                        if FRefRecord <> nil then
                            // it is the case, when records history was:
                            //                       [rsFetched]
                            //   [rsNewUpdated] <--> [rsOld]
                            //   [rsNewDeleted] <--> [rsOld]
                            SetFetched(pRH^.FRefRecord)
                        else
                            // in other cases they are:
                            //                       [rsFetched]
                            //                       [rsNewDeleted]
                            SetFetched(pRH);
                    end;
                end;
            if P = nil then
                Skip(1)
            else
                Break;
        end;
    finally
        FUpdatesPending := False;
        StatusFilter := prevStatusFilter;
        FDisableFetch := False;
        if P = nil then
            First(False)
        else if toP <> nil then
            SetCurrentRecord(toP, False);
    end;
end;

procedure TOCIDeadCursor.CommitUpdates(P: POCIBookmark);
var
    prevStatusFilter: TOCIRecordStatuses;
    lastPosBmk: TOCIDeadBookmark;
begin
    prevStatusFilter := StatusFilter;
    StatusFilter := [rsNewInsApplyed, rsNewUpdApplyed, rsNewDelApplyed];
    FDisableFetch := True;
    lastPosBmk := FPosition;
    try
        First(False);
        while not EOF do begin
            with FPosition, PageRecordHeader(FPageIndex, FRecIndex)^ do begin
                case FStatus of
                rsNewInsApplyed:
                    SetRecordStatus(FPageIndex, FRecIndex, rsFetched);
                rsNewDelApplyed:
                    begin
                        if FRefRecord <> nil then
                            FreeRecord(@FRefRecord^.FSelfBmk);
                        FreeRecord(@FSelfBmk);
                    end;
                rsNewUpdApplyed:
                    begin
                        SetRecordStatus(FPageIndex, FRecIndex, rsFetched);
                        if FRefRecord <> nil then
                            FreeRecord(@FRefRecord^.FSelfBmk);
                    end;
                end;
            end;
            Skip(1);
        end;
    finally
        FUpdatesPending := False;
        StatusFilter := prevStatusFilter;
        FDisableFetch := False;
        SetCurrentRecord(@lastPosBmk, not FInApplyUpdates);
    end;
end;

end.

